import { Accessor } from 'solid-js';
import { Context } from 'solid-js';
import { JSX } from 'solid-js';
import { JSXElement } from 'solid-js';
import { Result } from '@grinstead/result';

/**
 * A unique identifier for a binding within a pipeline.
 */
export declare type BindingId = {
    pipeline?: PipelineId;
    group: number;
    id: GPUIndex32;
};

/**
 * BufferBinding component binds an ArrayBufferView to a uniform in the shader.
 * @param props - The properties for the BufferBinding component.
 * @returns undefined
 *
 * @example
 * ```typescript
 * import { BufferBinding } from "@grinstead/webgpu";
 *
 * export function ExampleBufferBinding() {
 *   return (
 *     <BufferBinding
 *       pipeline="examplePipeline"
 *       group={0}
 *       id={0}
 *       label="Example Buffer"
 *       usage={GPUBufferUsage.UNIFORM | GPUBufferUsage.VERTEX}
 *       value={new Float32Array([0, 1, 2, 3])}
 *     />
 *   );
 * }
 * ```
 */
export declare function BufferBinding(props: BufferBindingProps): undefined;

/**
 * Props for the BufferBinding component.
 */
export declare type BufferBindingProps = BindingId & {
    /** Optional label for the buffer. */
    label?: string;
    /**
     * The usage, defaults to GPUBufferUsage.UNIFORM. It is not necessary to
     * include GPUBufferUsage.COPY_DST, that flag will always be injected in.
     */
    usage?: GPUBufferUsageFlags;
    /** The ArrayBufferView to bind. */
    value: ArrayBufferView;
};

/**
 * The simplest possible way to render a WebGPU Canvas.
 * This component creates a canvas element and initializes WebGPU.
 * It handles the device pixel ratio for high-DPI displays.
 *
 * @param props - The properties for the Canvas component.
 * @returns A SolidJS component that wraps a canvas element with WebGPU initialization.
 *
 * @example
 * ```typescript
 * import { Canvas } from "@grinstead/webgpu";
 *
 * export function ExampleCanvas() {
 *   return (
 *     <Canvas width={800} height={600}>
 *       ...WebGPU components can go here...
 *       <RenderPipeline />
 *     </Canvas>
 *   );
 * }
 * ```
 */
export declare function Canvas(props: CanvasProps): JSX.Element;

/**
 * Props for the Canvas component.
 */
export declare type CanvasProps = JSX.CanvasHTMLAttributes<HTMLCanvasElement>;

/**
 * Creates a binding for a given pipeline, group, and ID.
 * @param props - The binding ID properties.
 * @returns A function to update the binding resource.
 *
 * @example
 * ```typescript
 * import { createBinding } from "@grinstead/webgpu";
 *
 * const updateBinding = createBinding({ pipeline: "examplePipeline", group: 0, id: 0 });
 * updateBinding(resource);
 * ```
 */
export declare function createBinding(props: BindingId): Accessor<UpdateBinding>;

/**
 * Add code to call whenever the system renders.
 * @param code The code to render, dependencies are tracked
 *
 * @example
 * ```typescript
 * import { createGPURender } from "@grinstead/webgpu";
 *
 * createGPURender(() => (encoder) => {
 *   // GPU rendering commands go here
 * });
 * ```
 */
export declare function createGPURender(code: Accessor<GPUWork>): void;

/**
 * Add code to write to the GPU whenever the dependencies change.
 * @param code The code to render, dependencies are tracked
 *
 * @example
 * ```typescript
 * import { createGPUWrite } from "@grinstead/webgpu";
 *
 * createGPUWrite(() => (encoder) => {
 *   // GPU write commands go here
 * });
 * ```
 */
export declare function createGPUWrite(code: Accessor<GPUWork>): void;

/** Error code for when the computer has no acceptable gpu. */
export declare const E_GPU_NOHARDWARE = 404;

/**
 * Error code for when the browser does not supprot WebGPU.
 */
export declare const E_GPU_NOSUPPORT = 422;

/**
 * Error returned when the canvas fails to provide a WebGPU context,
 * could be because WebGPU is unsupported or (more likely) because
 * the canvas is already being used in a different way.
 */
export declare const E_GPU_NOWEBGPU = 423;

/**
 * Human-readable error messages for GPU loading failures.
 */
export declare const GPU_LOAD_ERROR: {
    readonly 422: "This browser does not support WebGPU";
    readonly 404: "This hardware is unsupported";
    readonly 423: "Failed to initialize";
};

/**
 * GPUContainer component to initialize and provide a WebGPU context.
 * @param props - The properties for the GPUContainer component.
 * @returns A SolidJS component that provides the WebGPU context.
 *
 * @example
 * ```typescript
 * import { GPUContainer } from "@grinstead/webgpu";
 *
 * export function ExampleGPUContainer() {
 *   const canvasRef = useRef<HTMLCanvasElement>(null);
 *
 *   return (
 *     <GPUContainer canvas={canvasRef.current}>
 *       WebGPU components can go here
 *     </GPUContainer>
 *   );
 * }
 * ```
 */
export declare function GPUContainer(props: GPUContainerProps): JSX.Element;

/**
 * Props for the GPUContainer component.
 */
export declare type GPUContainerProps = LoadGPUOptions & {
    /** The HTML canvas element to use for WebGPU rendering. */
    canvas: HTMLCanvasElement;
    /** An optional fallback element to display while the GPU context is being initialized. */
    fallback?: JSXElement;
    /**
     * An optional function that returns a JSX element to display if the GPU context fails to initialize.
     * If nothing is provided, a {@link GPULoadError} will be thrown that could be caught by a surrounding `ErrorBoundary`
     */
    failure?: (error: GPULoadErrorCode) => JSXElement;
    /**
     * The children elements that will be rendered inside the GPU context provider.
     */
    children: JSXElement;
};

/**
 * Context to provide GPUDetails.
 */
export declare const GPUContext: Context<GPUDetails>;

/**
 * Details about the GPU context.
 */
export declare type GPUDetails = {
    context: GPUCanvasContext;
    device: GPUDevice;
    canvas: HTMLCanvasElement;
    format: GPUTextureFormat;
};

/**
 * Error class for GPU loading failures.
 */
export declare class GPULoadError extends Error {
    readonly code: GPULoadErrorCode;
    constructor(code: GPULoadErrorCode);
}

/**
 * Type for GPU load error codes.
 */
export declare type GPULoadErrorCode = keyof typeof GPU_LOAD_ERROR;

/**
 * A function that performs work with a GPUCommandEncoder.
 */
export declare type GPUWork = (encoder: GPUCommandEncoder) => void;

/**
 * Class to manage GPU work queues, including persistent and one-time work.
 */
export declare class GPUWorkQueue {
    readonly gpu: GPUDetails;
    private readonly persistentWork;
    private readonly oneTimeWork;
    private readonly pipelines;
    hasWork: boolean;
    onHasWork: undefined | (() => void);
    private nextWorkId;
    static Provider(props: GPUWorkQueueProviderProps): JSX.Element;
    constructor(gpu: GPUDetails);
    /**
     * Reserves a slot for GPU work, allowing it to be updated.
     * @returns A function to update the work to be performed.
     */
    reserveSlot(): UpdateWork;
    /**
     * Creates a binding for a specific pipeline, group, and id.
     * @param pipelineId - The pipeline ID.
     * @param groupId - The group ID.
     * @param id - The binding ID.
     * @returns A function to update the binding resource.
     */
    makeBinding(pipelineId: PipelineId, groupId: number, id: GPUIndex32): (resource?: null | GPUBindingResource) => void;
    private _markWork;
    /**
     * Iterates over each bind group for a pipeline, applying a function to each.
     * @param pipelineId - The pipeline ID.
     * @param pipeline - The pipeline object.
     * @param map - The function to apply to each bind group.
     */
    forEachBindGroup(pipelineId: PipelineId, pipeline: GPUPipelineBase, map: (group: GPUBindGroup, id: number) => void): void;
    /**
     * Runs all queued GPU work.
     */
    runQueued(): void;
}

/**
 * Context for the GPUWorkQueue.
 */
export declare const GPUWorkQueueContext: Context<GPUWorkQueue | undefined>;

/**
 * Props for the GPUWorkQueueProvider.
 */
export declare type GPUWorkQueueProviderProps = {
    ref?: GPUWorkQueue | ((render: GPUWorkQueue) => void);
    onHasWork?: () => void;
    children: JSXElement;
};

/**
 * Loads and initializes the GPU context for a canvas element.
 * @param canvas - The canvas element to use for WebGPU.
 * @param options - Optional adapter and device options.
 * @returns A promise that resolves to a Result with GPUDetails or a GPULoadErrorCode.
 *
 * @example
 * ```typescript
 * import { loadGPU } from "@grinstead/webgpu";
 *
 * const canvas = document.createElement("canvas");
 * loadGPU(canvas).then((result) => {
 *   if (result.failure) {
 *     console.error("Failed to initialize WebGPU:", result.failure);
 *   } else {
 *     const { context, device, format } = result.value;
 *     // Use the GPU context...
 *   }
 * });
 * ```
 */
export declare function loadGPU(canvas: HTMLCanvasElement, options?: LoadGPUOptions): Promise<Result<GPUDetails, GPULoadErrorCode>>;

/**
 * Options for loading the GPU.
 */
export declare type LoadGPUOptions = {
    /** Optional adapter options for requesting a GPU adapter. */
    adapter?: GPURequestAdapterOptions;
    /** Optional device descriptor for requesting a GPU device. */
    device?: GPUDeviceDescriptor;
};

/**
 * A unique identifier for a pipeline.
 */
export declare type PipelineId = string | undefined;

/**
 * RenderPipeline component defines the render pipeline for WebGPU.
 * @param props - The properties for the RenderPipeline component.
 * @returns undefined
 *
 * @example
 * ```typescript
 * import { RenderPipeline, Canvas } from "@grinstead/webgpu";
 *
 * export function ExampleRenderPipeline() {
 *   return (
 *     <Canvas width={320} height={240}>
 *       <RenderPipeline
 *         label="Basic Render"
 *         vertexMain="vertexMain"
 *         fragmentMain="fragmentMain"
 *         draw={4}
 *         code={`
 * struct Vertex {
 *   @builtin(position) position: vec4f,
 *   @location(0) fragUV: vec2f,
 * }
 *
 * @vertex
 * fn vertexMain(@builtin(vertex_index) index: u32) -> Vertex {
 *   const fragPoints = array(
 *     vec2f(0, 0),
 *     vec2f(1, 0),
 *     vec2f(0, 1),
 *     vec2f(1, 1),
 *   );
 *
 *   let uv = fragPoints[index];
 *
 *   return Vertex(
 *     vec4f(2 * uv - 1, 0, 1),
 *     uv,
 *   );
 * }
 *
 * @fragment
 * fn fragmentMain(@location(0) uv: vec2f) -> @location(0) vec4f {
 *     return vec4f(1, uv.x, uv.y, 1);
 * }
 * `}
 *       />
 *     </Canvas>
 *   );
 * }
 * ```
 */
export declare function RenderPipeline(props: RenderPipelineProps): undefined;

/**
 * Props for the RenderPipeline component.
 */
export declare type RenderPipelineProps = {
    /**
     * Optional pipeline ID for the bind groups.
     */
    pipeline?: PipelineId;
    /**
     * Optional label for the pipeline.
     */
    label?: string;
    /**
     * The WGSL code for the shaders.
     */
    code: string;
    /**
     * The entry point for the vertex shader.
     */
    vertexMain: string;
    /**
     * The entry point for the fragment shader.
     */
    fragmentMain: string;
    /**
     * The number of vertices to draw or a function to handle custom drawing.
     */
    draw: number | ((encoder: GPURenderPassEncoder) => void);
    /**
     * Optional color attachments for the render pass.
     */
    colorAttachments?: Iterable<GPURenderPassColorAttachment | null>;
};

/**
 * ScalarBinding component binds a scalar value to a uniform in the shader.
 * @param props - The properties for the ScalarBinding component.
 * @returns undefined
 *
 * @example
 * ```typescript
 * import { ScalarBinding } from "@grinstead/webgpu";
 * import { createSignal } from "solid-js";
 *
 * export function ExampleScalarBinding() {
 *   const [red, setRed] = createSignal(0);
 *
 *   return (
 *     <Canvas width={320} height={240} onClick={() => setRed(Math.random())}>
 *       <RenderPipeline
 *         label="Scalar Binding Example"
 *         vertexMain="vertexMain"
 *         fragmentMain="fragmentMain"
 *         draw={4}
 *         code={`
 * struct Vertex {
 *   @builtin(position) position: vec4f,
 *   @location(0) fragUV: vec2f,
 * }
 *
 * @vertex
 * fn vertexMain(@builtin(vertex_index) index: u32) -> Vertex {
 *   const fragPoints = array(
 *     vec2f(0, 0),
 *     vec2f(1, 0),
 *     vec2f(0, 1),
 *     vec2f(1, 1),
 *   );
 *
 *   let uv = fragPoints[index];
 *
 *   return Vertex(
 *     vec4f(2 * uv - 1, 0, 1),
 *     uv,
 *   );
 * }
 *
 * @group(0) @binding(0) var<uniform> red: f32;
 *
 * @fragment
 * fn fragmentMain(@location(0) uv: vec2f) -> @location(0) vec4f {
 *     return vec4f(red, uv.x, uv.y, 1);
 * }
 * `}
 *       />
 *       <ScalarBinding group={0} id={0} value={red()} />
 *     </Canvas>
 *   );
 * }
 * ```
 */
export declare function ScalarBinding(props: ScalarBindingProps): undefined;

/**
 * Props for the ScalarBinding component.
 */
export declare type ScalarBindingProps = BindingId & {
    /**
     * Optional label for the buffer.
     */
    label?: string;
    /**
     * The usage, defaults to GPUBufferUsage.UNIFORM. It is not necessary to
     * include GPUBufferUsage.COPY_DST, that flag will always be injected in.
     */
    usage?: GPUBufferUsageFlags;
    /**
     * Data type for the scalar value. Defaults to 'f32'.
     */
    datatype?: WebGPUScalar;
    /**
     * The scalar value to bind.
     */
    value: number;
};

/**
 * All uniform vriables must be passed in on 16-byte alignment
 */
export declare const UNIFORM_ALIGNMENT = 16;

/**
 * A function that updates a binding resource.
 */
export declare type UpdateBinding = (resource: null | GPUBindingResource) => void;

/**
 * A function that updates the work to be performed, optionally on every frame.
 */
export declare type UpdateWork = (update: null | GPUWork, everyFrame?: boolean) => void;

/**
 * VectorBinding component binds a vector value to a uniform in the shader.
 * @param props - The properties for the VectorBinding component.
 * @returns undefined
 *
 * @example
 * ```typescript
 * import { VectorBinding } from "@grinstead/webgpu";
 * import { createSignal } from "solid-js";
 *
 * export function ExampleVectorBinding() {
 *   const [color, setColor] = createSignal([0, 0, 1, 1]);
 *
 *   return (
 *     <Canvas width={320} height={240} onClick={() => setColor([Math.random(), Math.random(), Math.random(), 1])}>
 *       <RenderPipeline
 *         label="Vector Binding Example"
 *         vertexMain="vertexMain"
 *         fragmentMain="fragmentMain"
 *         draw={4}
 *         code={`
 * struct Vertex {
 *   @builtin(position) position: vec4f,
 *   @location(0) fragUV: vec2f,
 * }
 *
 * @vertex
 * fn vertexMain(@builtin(vertex_index) index: u32) -> Vertex {
 *   const fragPoints = array(
 *     vec2f(0, 0),
 *     vec2f(1, 0),
 *     vec2f(0, 1),
 *     vec2f(1, 1),
 *   );
 *
 *   let uv = fragPoints[index];
 *
 *   return Vertex(
 *     vec4f(2 * uv - 1, 0, 1),
 *     uv,
 *   );
 * }
 *
 * @group(0) @binding(0) var<uniform> color: vec4f;
 *
 * @fragment
 * fn fragmentMain(@location(0) uv: vec2f) -> @location(0) vec4f {
 *     return vec4f(color.r, color.g, color.b, 1);
 * }
 * `}
 *       />
 *       <VectorBinding group={0} id={0} value={color()} />
 *     </Canvas>
 *   );
 * }
 * ```
 */
export declare function VectorBinding(props: VectorBindingProps): undefined;

/**
 * Props for the VectorBinding component.
 */
export declare type VectorBindingProps = BindingId & {
    /**
     * Optional label for the buffer.
     */
    label?: string;
    /**
     * The usage, defaults to GPUBufferUsage.UNIFORM. It is not necessary to
     * include GPUBufferUsage.COPY_DST, that flag will always be injected in.
     */
    usage?: GPUBufferUsageFlags;
    /**
     * Data type for the vector values. Defaults to 'f32'.
     */
    datatype?: WebGPUScalar;
    /**
     * The vector values to bind.
     * Can be a 2, 3, or 4 element tuple.
     */
    value: [number, number] | [number, number, number] | [number, number, number, number];
};

export declare type WebGPUScalar = "u32" | "i32" | "f32";

export { }
